/*
 Illinois Open Source License
 
 University of Illinois/NCSA
 Open Source License
 
 Copyright © 2009,    University of Illinois.  All rights reserved.
 
 Developed by: 
 Innovative Systems Lab
 National Center for Supercomputing Applications
 http://www.ncsa.uiuc.edu/AboutUs/Directorates/ISL.html
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal with the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimers.
 
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimers in the documentation and/or other materials provided with the distribution.
 
 * Neither the names of Innovative Systems Lab and National Center for Supercomputing Applications, nor the names of its contributors may be used to endorse or promote products derived from this Software without specific prior written permission.
 
 THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
 */

#include <sys/time.h>
#include <string.h>
#include <math.h>

#include "model.h"

int doComputeBSEARC(struct cartesian *data1, int n1, struct cartesian *data2, int n2, 
              int doSelf, long long **data_bins, int nbins, double *binb, int njk)
{
    int i, j, l;
    
    if (doSelf)
    {
        n2 = n1;
        data2 = data1;
    }

    for (i = 0; i < ((doSelf) ? n1-1 : n1); i++)
    {
        const register double xi = data1[i].x;
        const register double yi = data1[i].y;
        const register double zi = data1[i].z;
        const register int jk = data1[i].jk;

        for (j = ((doSelf) ? i+1 : 0); j < n2; j++)
        {
            register double dot = xi * data2[j].x + yi * data2[j].y + zi * data2[j].z;

            // BINARY SEARCH
            register int indx;
            register int min = 0;
            register int max = nbins;

            if (dot >= binb[min])
                indx = min;
            else if (dot < binb[max])
                indx = max + 1;
            else 
            {
                register int k;

                // run binary search
                while (max > min+1)
                {
                    k = (min + max) / 2;
                    if (dot >= binb[k]) 
                        max = k;
                    else 
                        min = k;
                };
                
                indx = max;
            }

            // update all samples
            for (l = 0; l < njk; l++)
            {
                if (l != jk) data_bins[l][indx] += 1;
            }
        }
    }

    return 0;
}




int doComputeWATERFALL(struct cartesian *data1, int n1, struct cartesian *data2, int n2, 
              int doSelf, long long **data_bins, int nbins, double *binb, int njk)
{
    int i, j, l;
    
    if (doSelf)
    {
        n2 = n1;
        data2 = data1;
    }

    for (i = 0; i < ((doSelf) ? n1-1 : n1); i++)
    {
        const register double xi = data1[i].x;
        const register double yi = data1[i].y;
        const register double zi = data1[i].z;
        const register int jk = data1[i].jk;

        for (j = ((doSelf) ? i+1 : 0); j < n2; j++)
        {
            register int indx;
            register double dot = xi * data2[j].x + yi * data2[j].y + zi * data2[j].z;

            // WATERFALL
            register int k = nbins; 
            if (dot >= binb[0])
                indx = 0; // data_bins[0] += 1;
            else
            {
                while (dot > binb[k]) k--;
                indx = k+1; // data_bins[k+1] += 1;
            }

            // update all samples
            for (l = 0; l < njk; l++)
            {
                if (l != jk) data_bins[l][indx] += 1;
            }
        }
    }

    return 0;
}

